; ==================================================================
; MULTFLP multipliziert 2 positive FLP-Binrzahlen mit einer
; Laenge von 8+16 bit. Das Ergebnis der Multiplikation
; wird gerundet und als FLP 8+16-bit-Zahl ausgegeben.
; Vorgang: (y) * (z) = (z), Adressen in y und z werden nicht verndert
; die Zahl in (y) wird nicht verndert.
; Zeit: ca. 33us bei 8MHz

FLPMULTIPL:
; MULTIPLIKATION DER MANTISSEN
	clr	r2
	clr	r3
	ldd	acc,y+2		;ARITIN+2
	ldd	acc2,z+2	;ARITOUT+2
	mul	acc,acc2
	
	mov	r5,r0
	mov	r4,r1
	ldd	acc,y+1		;ARITIN+1
	ldd	acc2,z+2	;ARITOUT+2
	mul	acc,acc2
	
	add	r4,r0
	adc	r3,r1
	ldi	acc,0
	adc	r2,acc
	ldd	acc,y+2		;ARITIN+2
	ldd	acc2,z+1	;ARITOUT+1
	mul	acc,acc2
	
	add	r4,r0
	adc	r3,r1
	ldi	acc,0
	adc	r2,acc
	ldd	acc,y+1		;ARITIN+1
	ldd	acc2,z+1	;ARITOUT+1
	mul	acc,acc2

	add	r3,r0
	adc	r2,r1
; ADDITION DER EXPONENTEN
	ld	acc,y		;ARITIN
	ld	acc2,z		;aritout
	add	acc,acc2
	subi	acc,-16		; KORREKTUR WEGEN STELLENVERLUST
	st	z,acc		;ARITOUT,acc
; ERGEBNIS SCHIEBEN BIS HW-BIT = 1 IST
	mov	acc,r2
	sbrc	acc,7
	rjmp	MULT3
	ldi	acc,31
	mov	R6,acc
MULT1:
	ld	acc,z		;aritout
	dec	acc
	st	z,acc		;aritout,acc
	clc
	rol	r5
	rol	r4
	rol	r3
	rol	r2
	dec	r6
	brne	MULT11
	rjmp	MULT3
MULT11:
	sbrs	r2,7
	rjmp    MULT1

MULT3:	; ERGEBNIS AUFRRUNDEN

	sbrs	r4,7
	rjmp	MULT4		; AUFRUNDEN UNNOETIG!
	sec
	ldi	acc,0
	adc	r3,acc
	adc	r2,acc
	brcc	MULT4		; UEBERTRAG?
	ror	r2
	ror	r3
	ld	acc,z		;aritout
	inc	acc
	st	z,acc		;ARITOUT,acc
MULT4: 
	std	z+1,r2		;ARITOUT+1,r2
	std	z+2,r3		;ARITOUT+2,r3
	ret

; =========================================================================
; KEHRWERT bildet den Kehrwert aus einer positiven FLP-
; Binrzahl. Mantisse = 16 bit, Exponent = 8 bit.
; Eingang: (y)
; Ausgang: (y), Adresse in y wird nicht verndert
; Zeit: ca. 52us bei 8MHz

KEHRWERT:
	ldi	lcdzeiger,16		; SCHLEIFENZAEHLER

; DIVISOR IN REGISTER LADEN

	ldd	R3,y+1	; NW-BYTE DIVISOR
	ldd	R4,y+2	; HW-BYTE
KWRT1:	; DIVISOR SCHIEBEN BIS HW-BIT=1 IST
	clc
	rol	r4
	rol	r3
	brcs	KWRT2		; HOECHSTES BIT IN C?
	ld	acc,y
	dec	acc
	st	y,acc
	dec	lcdzeiger
	brne    KWRT1
	sec			; DIVISOR=0!!
	ret
KWRT2: 
	ror	r3
	ror	r4
	ldi	lcdzeiger,16		; SCHLEIFENZAEHLER
	clr	R7		; DIVIDENT NW
	ldi	acc,0x80
	mov	R6,acc		; DIVIDENT HW
	clr	R5		; DIVIDENT UEBERLAUF

KWRT3:   ;:DIVISION DER MANTISSEN

	sub	r7,R4
	sbc	r6,R3
	ldi	acc,0
	sbc	r5,acc
	brcc	KWRT4		; DIVIDENT>DIVISOR?JA
	add	r7,R4
	adc	r6,R3
	adc	r5,acc
	clc
	rjmp	KWRT5
KWRT4: 
	sec
KWRT5: 
	rol	r1
	rol	r0

; DIVIDENT SCHIEBEN

	clc
	rol	r7
	rol	r6
	rol	r5
	dec	lcdzeiger
	brne	KWRT3

;:KORREKTUR DER EXPONENTEN

	ldi	acc,-15		; EXP DES DIVIDENTEN
	ld	acc2,y
	sub	acc,acc2
	subi	acc,15		; VERSCHIEBUNG DES DIVISORS
	st	y,acc

;:ERGEBNIS AUFRUNDEN

	mov	acc,R7
	sub	acc,R4
	mov	acc,R6
	sbc	acc,R3
	mov	acc,R5
	sbci	acc,0
	brcs	KWRT6		; KEIN AUFRUNDEN?
	sec
	ldi	acc,0
	adc	R1,acc
	adc	R0,acc
	brcc	KWRT6
	ror	r0
	ror	r1
	ld	acc,y
	inc	acc		; EXPONENT+1
	st	y,acc

KWRT6:	; ERGEBNIS IN DEN INTERNEN RAM LADEN

	clc
	std	y+1,R0
	std	y+2,R1
	ret

; ======================================================================
; EXPKORR bringt die FLP-Zahl in (z) auf den
; in acc uebergebenen Exponenten. Wenn beim Linksschieben
; Stellen verloren gehen, wird T-flag gesetzt.
; Zeit: ca. 9us bei 8MHz

EXPKORR: 
	clt
	push	acc
	;:MANTISSE -> R4/R5
	ldd	r4,z+1
	ldd	r5,z+2
	clc
	ld	acc2,z		; Adr Exponent
	sub	acc,acc2
	breq	EXPRET		; Exp = 0!
	sbrc	acc,7
	rjmp	EXPK1		; NEGATIVE KORREKTUR?
	mov	R3,acc
EXPK0:  ;:MANTISSE -->
	clc
	ror	r4
	ror	r5
	dec	r3
	brne	EXPK0
	brcc	EXPK3
; AUFRUNDEN
	ldi	acc,1
	add	R5,acc
	ldi	acc,0
	adc	r4,acc
	rjmp	EXPK3
EXPK1: 
	com	acc
	inc	acc
	mov	R3,acc
EXPK2:	; MANTISSE <--
	clc
	rol	r5
	rol	r4
	brcc	EXPK21		; KEIN UEBERLAUF?
	set
EXPK21:
	dec	r3
	brne	EXPK2
EXPK3: 
	pop	acc
	st	z,acc
	std	z+1,R4
	std	z+2,R5
	ret
EXPRET:
	pop	acc
	ret

; FLPADD addiert 2 FLP-Zahlen im Format 1+2-Byte
; (y) + (z) = (z), Adressen in y und z werden nicht verndert
flpadd:
	ld	r2,y
	ldd	r3,y+1
	ldd	r4,y+2
	ld	r5,z
	ldd	r6,z+1
	ldd	r7,z+2
;	sbiw	zh:zl,3
	ldi	acc,0x7f
	add	r2,acc		; Exp ins Positive Drehen
	add	r5,acc
	mov	r1,r5
	cp	r2,r5
	breq	flpadd5		; beide haben den selben Exponenten
	brcs	flpadd2		; Exponent in r5 ist grer
; Exponent in r2 ist grer, r5 mu grer = rechts schieben!
	mov	r1,r2		; Ausgangs-Exp sichern
	sub	r2,r5
	mov	lcdzeiger,r2
flpadd1:	
	lsr	r6
	ror	r7
	dec	lcdzeiger
	brne	flpadd1
	rjmp	flpadd5
; Exponent in r5 ist grer, r2 mu grer = rechts schieben!
flpadd2:	
	mov	r1,r5		; Ausgangs-Exp sichern
	sub	r5,r2
	mov	lcdzeiger,r5
flpadd3:
	lsr	r3
	ror	r4
	dec	lcdzeiger
	brne	flpadd3
flpadd5:
	add	r7,r4
	adc	r6,r3
	brcc	flpadd6
; berlauf: C,r6,r7 1x nach rechts schieben und Exponent +1	
	ror	r6
	ror	r7
	inc	r1	
flpadd6:
	ldi	acc,0x7f
	sub	r1,acc
	st	z,r1
	std	z+1,r6
	std	z+2,r7
	ret
	
